@@ -5,9 +5,10 @@ import typing
 from ccxt.base import errors as ccxt_errors
 import ccxt.pro as ccxt
 from contextlib import suppress
+from ccxt import DECIMAL_PLACES
 
 # import ccxt.async_support as ccxt
-from decimal import Decimal
+from decimal import Decimal, ROUND_DOWN
 from datetime import datetime, timedelta, timezone
 from tortoise import connections
 import json
@@ -47,7 +48,7 @@ class ZCCXT:
         None  # unable to set ccxt.Exchange type because all methods in that base class raise NotImplemented exception
     )
 
-    def __init__(self, api_key: str, api_secret: str, exchange: ExchangeEnum, is_demo=False, identifier=None):
+    def __init__(self, api_key: str, api_secret: str, exchange: ExchangeEnum, is_demo=False, identifier=None, disable_db=False):
         if exchange.value not in [e.value for e in ExchangeEnum]:
             raise ECInvalidValue(f"Invalid exchange provided: {exchange}")
 
@@ -56,10 +57,11 @@ class ZCCXT:
         self.exchange = exchange
         self.is_demo = is_demo
         self.identifier = identifier
+        self.db_disabled = disable_db
 
     @classmethod
-    async def create(cls, api_key: str, api_secret: str, exchange: ExchangeEnum, is_demo=False, identifier=None):
-        self = cls(api_key, api_secret, exchange, is_demo, identifier)
+    async def create(cls, api_key: str, api_secret: str, exchange: ExchangeEnum, is_demo=False, identifier=None, disable_db=False):
+        self = cls(api_key, api_secret, exchange, is_demo, identifier, disable_db)
 
         params = {
             "apiKey": self.api_key,
@@ -83,12 +85,15 @@ class ZCCXT:
             params.update({"uid": self.identifier})
 
         self.client = getattr(ccxt, self.exchange.lower())(params)
-
-        conn = connections.get("default")
-        _, res = await conn.execute_query(
-            "SELECT * FROM ccxt_market_cache WHERE exchange = $1",
-            (self.exchange.value,),
-        )
+        
+        if not self.db_disabled:
+            conn = connections.get("default")
+            _, res = await conn.execute_query(
+                "SELECT * FROM ccxt_market_cache WHERE exchange = $1",
+                (self.exchange.value,),
+            )
+        else:
+            res = []
 
         if len(res) > 0:
             key = res[0]
@@ -107,6 +112,8 @@ class ZCCXT:
         return self
 
     async def cache_data(self, data):
+        if self.db_disabled:
+            return
         conn = connections.get("default")
         await conn.execute_query(
             """
@@ -966,3 +973,25 @@ class ZCCXT:
             elif "order has been terminated" in e.args[0].lower():  # WOO error ...
                 return
             raise e
+
+    def round_qty(self, symbol: str, qty):
+        try:
+            proposed_qty = float(abs(qty))
+            qty = float(self.client.amount_to_precision(symbol, proposed_qty))
+            # amount_to_precision does not round numbers down but it rounds it up, causing issue when selling SPOT
+            # because we want to sell more than we have...
+            # TIL: float(0.9) > Decimal("0.9") = True ... this is because float(0.9)  # ≈ 0.900000000000000022
+            if Decimal(str(qty)) > proposed_qty:
+                amount_precision = Decimal(str(self.client.markets[symbol]["precision"]["amount"]))
+                # if amount_precision > 0:
+                #     amount_precision = int(amount_precision)
+                if self.client.precisionMode == DECIMAL_PLACES:
+                    amount_precision = Decimal("10") ** -amount_precision
+                qty = Decimal(proposed_qty).quantize(amount_precision, rounding=ROUND_DOWN)
+        except ccxt_errors.InvalidOrder:
+            qty = 0.0
+
+        return qty
+
+    def round_price(self, symbol, price):
+        return self.client.price_to_precision(symbol, float(price))
